"""Tools for the consumenten prijs index (CPI)"""

__all__ = [
    "get_CPI_series",
]

import cbsodata
from datetime import datetime
import pandas as pd
from typing import Literal

import triodos_fund_analyzer as tfa


def _download_index(index_acro: Literal["CPI", "HICP"] = "CPI") -> None:
    if index_acro not in ["CPI", "HICP"]:
        raise Exception(f"Unexpected index ID {index_acro}.")
    pd.DataFrame(
        cbsodata.get_data(f"8313{1 if index_acro == 'CPI' else 3}ENG")
    ).set_index("ID").to_csv(tfa.data_path / f"{index_acro}.csv")


def _extract_main_series(df: pd.DataFrame) -> pd.Series:
    #  select "all items" (NL: alle bestedingen)
    df = df.loc[df.iloc(1)[0].apply(lambda x: x.startswith("000000"))].iloc(1)[1:]
    df = df.loc[df.iloc(1)[0].apply(lambda x: len(x) > 4)]  # drop yearly values
    df.index = pd.to_datetime(df.iloc(1)[0], format="%Y %B")
    return df.iloc(1)[-1]


def get_CPI_series(harmonized: bool = False) -> pd.Series:
    _abbr = "CPI" if not harmonized else "HICP"
    path = (tfa.data_path / _abbr).with_suffix(".csv")
    if (
        not path.exists()
        or datetime.now().month > datetime.fromtimestamp(path.stat().st_mtime).month
    ):
        _download_index(_abbr)
    return _extract_main_series(pd.read_csv(path, usecols=[1, 2, 3]))



"""Tool to visualize Triodos fund histories"""

__version__ = "0.1.dev1"

__all__ = [
    "CPI",
    "data_path",
    "funds",
    "plot_history",
    "cli",
]

import matplotlib.pyplot as plt
import pandas as pd
from pathlib import Path
from typing import Literal

from triodos_fund_analyzer.CPI import get_CPI_series
from triodos_fund_analyzer import funds

data_path = Path(__file__).parent / ".cached_data"
if not data_path.exists():
    data_path.mkdir()


def plot_history(
    fund: str,
    *,
    normalize: bool | float = 100.0,
    purchasing_power: bool | Literal["CPI", "HICP"] = "CPI",
    deduce_costs: bool = True,
    consider_dividends: bool = True,
    start_date: str = None,
    add_baseline: bool = False,
    ax: plt.Axes = None,
) -> plt.Axes:
    trading_value_history = funds.get_trading_value_series(fund)
    if start_date is not None:
        trading_value_history = trading_value_history.loc[start_date:]
    if normalize:
        trading_value_history /= trading_value_history.iloc[0] / (
            normalize if not isinstance(normalize, bool) else 100
        )
    if purchasing_power:
        index = get_CPI_series(harmonized=(purchasing_power == "HICP")).reindex(
            trading_value_history.index, method="ffill"
        )
        trading_value_history /= index / index.iloc[-1]
    relative_dividends = (funds.get_dividends(fund) / trading_value_history).fillna(0)
    fund_details = funds.get_details(fund)
    if deduce_costs:
        trading_value_history *= 0.991 if funds.get_acro(fund) == "TETEF" else 0.996
        relative_costs = pd.Series(
            fund_details.loc[["fond_costs", "service_fee", "transaction_fee"]].sum()
            * trading_value_history.index.diff()
            / pd.Timedelta(365.25, "days"),
            trading_value_history.index,
        )
        if consider_dividends:
            shares = 0.996 * (1 + relative_dividends - relative_costs).cumprod()  #
        else:
            shares = 0.996 * (1 - relative_costs).cumprod()
    else:
        if consider_dividends:
            shares = 1 * (1 + relative_dividends).cumprod()
        else:
            shares = 1
    if ax is None:
        _, ax = plt.subplots()
    line = (
        (shares * trading_value_history)
        .plot(ax=ax, label=funds.get_name(fund))
        .get_lines()[-1]
    )
    if add_baseline:
        baseline_properties = dict(
            color=line.get_c(),
            linewidth=line.get_lw() * 0.67,
            linestyle="dashed",
            zorder=line.zorder - 1,
            label="",
        )
        if purchasing_power:
            (trading_value_history.iloc[0] / (index / index.iloc[0])).plot(
                ax=ax, **baseline_properties
            )
        else:
            ax.hlines(
                trading_value_history.iloc[0],
                *trading_value_history.index[[0, -1]],
                **baseline_properties,
            )
    return ax


def cli():
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="show_fund", description="Visualize fund histories")
    parser.add_argument(
        "--normalize",
        nargs=1,
        default=[100],
        type=float,
        help="Set 0 to disable normalization",
        metavar="12.3",
    )
    parser.add_argument(
        "--purchasing-power",
        nargs=1,
        default=["CPI"],
        type=str,
        choices=["CPI", "HICP", "off"],
        help="Choose index to account for inflation",
    )
    parser.add_argument(
        "--ignore-costs",
        action="store_true",
        help="Do not reduce fund value by fund costs and service fees",
    )
    parser.add_argument(
        "--ignore-dividends",
        action="store_true",
        help="Do not add/reinvest dividends to fund value",
    )
    parser.add_argument(
        "--start",
        type=str,
        help="Start the history at provided date",
        metavar="yyyy-mm-dd",
    )
    parser.add_argument(
        "--hide-baseline",
        action="store_true",
        help="Do not show reference purchasing power evolution (will be hidden by "
        "default for 4 shown funds or more)",
    )
    parser.add_argument(
        "funds", nargs="+", help="List funds that should be shown", metavar="fund name"
    )
    args = parser.parse_args()

    options = dict(
        normalize=args.normalize[0] if args.normalize[0] > 0 else False,
        purchasing_power=(
            args.purchasing_power[0] if args.purchasing_power[0] != "off" else False
        ),
        deduce_costs=not args.ignore_costs,
        consider_dividends=not args.ignore_dividends,
        start_date=None if "start" not in args else args.start,
        add_baseline=not args.hide_baseline or len(args.funds) > 3,
    )

    _, ax = plt.subplots()
    for fund in args.funds:
        plot_history(fund, ax=ax, **options)
    ax.set_xlabel("")
    plt.legend()
    plt.ion()
    plt.show(block=True)



"""Browser app"""

import matplotlib.pyplot as plt
import streamlit as st

import triodos_fund_analyzer as tfa
from triodos_fund_analyzer.funds import details_df


def build_page():
    st.title("Triodos Fund Analyzer")
    fig, ax = plt.subplots()
    ax.clear()
    options = dict(
        normalize=st.session_state["normalize"]
        if st.session_state["normalize"] > 0
        else False,
        purchasing_power=(
            st.session_state["purchasing_power"]
            if st.session_state["purchasing_power"] != "off"
            else False
        ),
        deduce_costs=st.session_state["deduce_costs"],
        consider_dividends=st.session_state["consider_dividends"],
        start_date=st.session_state["start_date"],
        add_baseline=st.session_state["add_baseline"],
    )
    for fund in st.session_state["funds"]:
        tfa.plot_history(fund, ax=ax, **options)
    ax.set_xlabel("")
    ax.legend()
    fig.tight_layout()
    plt.ion()
    plt.show()
    st.pyplot(fig=fig, clear_figure=True, use_container_width=True)


if __name__ == "__main__":
    st.sidebar.segmented_control(
        "Funds",
        details_df.index.values,
        selection_mode="multi",
        key="funds",
        help=details_df.full_name.to_string(),
    )
    st.sidebar.slider("normalize", 0, 100, 100, key="normalize")
    st.sidebar.segmented_control(
        "purchasing power",
        ["off", "CPI", "HICP"],
        default="CPI",
        help="Choose index to account for inflation",
        key="purchasing_power",
    )
    st.sidebar.toggle(
        "deduce costs",
        True,
        help="Reduce fund value by fund costs and service fees",
        key="deduce_costs",
    )
    st.sidebar.toggle("consider dividends", True, key="consider_dividends")
    st.sidebar.date_input("start date", "1990-01-01", key="start_date")
    st.sidebar.toggle("add baseline", True, key="add_baseline")

    build_page()



"""Triodos fund module"""

__all__ = [
    "details_df",
]

import pandas as pd
from pathlib import Path
import requests

import triodos_fund_analyzer as tfa

details_df = pd.DataFrame.from_dict(
    {
        "TGF": {
            "unique_id": "NL0000440204",
            "full_name": "Triodos Groenfonds",
            "fond_costs": 0.0097,
            "service_fee": 0.0048,
            "transaction_fee": 0.0,
        },
        "TFSF": {
            "unique_id": "NL0013087968",
            "full_name": "Triodos Fair Share Fund",
            "fond_costs": 0.024,
            "service_fee": 0.0048,
            "transaction_fee": 0.0,
        },
        "TETEF": {
            "unique_id": "NL0013908700",
            "full_name": "Triodos Energy Transition Europe Fund",
            "fond_costs": 0.0258,
            "service_fee": 0.0048,
            "transaction_fee": 0.0005,
        },
        "TMIF": {
            "unique_id": "NL00150022X1",
            "full_name": "Triodos Multi Impact Fund",
            "fond_costs": 0.02,
            "service_fee": 0.0048,
            "transaction_fee": 0.0,
        },
        "TEBIF": {
            "unique_id": "LU0785617936",
            "full_name": "Triodos Euro Bond Impact Fund",
            "fond_costs": 0.0065,
            "service_fee": 0.0048,
            "transaction_fee": 0.0,
        },
        "TIMFD": {
            "unique_id": "LU1956011438",
            "full_name": "Triodos Impact Mixed Fund - Defensive",
            "fond_costs": 0.0085,
            "service_fee": 0.0048,
            "transaction_fee": 0.0001,
        },
        "TIMFN": {
            "unique_id": "LU0785618405",
            "full_name": "Triodos Impact Mixed Fund - Neutral",
            "fond_costs": 0.009,
            "service_fee": 0.0048,
            "transaction_fee": 0.0002,
        },
        "TIMFO": {
            "unique_id": "LU1956012089",
            "full_name": "Triodos Impact Mixed Fund - Offensive",
            "fond_costs": 0.0095,
            "service_fee": 0.0048,
            "transaction_fee": 0.0002,
        },
        "TPIF": {
            "unique_id": "LU0785618744",
            "full_name": "Triodos Pioneer Impact Fund",
            "fond_costs": 0.011,
            "service_fee": 0.0048,
            "transaction_fee": 0.0004,
        },
        "TFGF": {
            "unique_id": "LU2434354713",
            "full_name": "Triodos Future Generations Fund",
            "fond_costs": 0.011,
            "service_fee": 0.0048,
            "transaction_fee": 0.0003,
        },
        "TGEIF": {
            "unique_id": "LU0785617423",
            "full_name": "Triodos Global Equities Impact Fund",
            "fond_costs": 0.01,
            "service_fee": 0.0048,
            "transaction_fee": 0.0003,
        },
    },
    orient="index",
)


def _download_history(fund: str) -> None:
    details = get_details(fund)
    resp = requests.get(
        "https://www.triodos.nl/fund-data-download"
        f"?fund={details.name}"
        f"&isin={details.unique_id}"
        "&price=TRADING_SHARE_PRICE",
        params={"format": "xlsx"},
    )
    if not resp.ok:
        Exception(
            "Error occurred while downloading the data. If this persists, please file "
            f"an issue.\nFailing request URL:\n{resp}"
        )
    with open(get_data_path(fund), "wb") as f:
        f.write(resp.content)


def get_details(fund: str) -> pd.Series:
    if fund in details_df.index:
        return details_df.loc[fund]
    elif fund in details_df.unique_id.values:
        return details_df[details_df.unique_id == fund].iloc[0]
    elif fund in details_df.full_name.values:
        return details_df[details_df.full_name == fund].iloc[0]
    else:
        raise Exception(f"Not a valid fund descriptor: {fund}")


def get_acro(fund: str) -> str:
    return get_details(fund).name


def get_id(fund: str) -> str:
    return get_details(fund).unique_id


def get_name(fund: str) -> str:
    return get_details(fund).full_name


def get_data_path(fund: str) -> Path:
    details = get_details(fund)
    return (tfa.data_path / "-".join([details.name, details.unique_id])).with_suffix(
        ".xlsx"
    )


def get_dividends(fund: str) -> pd.Series:
    if "Dividend History" not in pd.ExcelFile(get_data_path(fund)).sheet_names:
        return 0 * get_trading_value_series(fund).iloc[[-1]]
    return (
        pd.read_excel(
            get_data_path(fund),
            "Dividend History",
            header=7,
            usecols=[2, 3],
            index_col=0,
        )
        .iloc(1)[0]
        .sort_index()
    )


def get_trading_value_series(fund: str) -> pd.Series:
    path = get_data_path(fund)
    if not path.exists():
        _download_history(fund)
    return (
        pd.read_excel(
            path,
            "Historical NAV",
            header=10,
            usecols=[1, 2],
            index_col=0,
        )
        .iloc(1)[0]
        .sort_index()
    )
